from .. import *
from ..core.async_sync import *
from .get_type import *
from .props import *
from .models.chat import Chat
from .metadata import metadata as metadata_prop
from .file import File
from ..utils.utils import Utils
import json
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union, Literal
import re

if TYPE_CHECKING:
    from .prop_update import msg_update
    from ..core import Client

class Update:
    def __init__(self, update_data: dict, client: "Client"):
        try:
            self._data = update_data["new_message"]
        except:
            self._data = update_data["updated_message"]
        self._client = client
        self.raw_data_ = update_data
    @property
    def text(self) -> Optional[str]:
        """text message / متن پیام"""
        return self._data.get("text")
    @property
    def message_id(self) -> str:
        """message id / آیدی پیام"""
        return self._data['message_id']
    @property
    def chat_id(self) -> str:
        """chat id message / چت آیدی پیام"""
        return self.raw_data_['chat_id']
    @property
    def time(self) -> int:
        """time sended message / زمان ارسال شده پیام"""
        return int(self._data['time'])
    @property
    def sender_type(self) -> Literal["User","Group","Channel"]:
        """sender type / نوع ارسال کننده"""
        return Utils.get_chat_id_type(self.chat_id)
    @property
    def sender_id(self) -> str:
        """sender id message / شناسه گوید کاربر ارسال کننده"""
        return self._data['sender_id']
    @property
    def is_edited(self):
        return self._data['is_edited']
    
    # File

    @property
    def file(self) -> Optional[dict]:
        """file / فایل"""
        return self._data['file'] if "file" in self._data else None
    @property
    def file_id(self) -> Optional[str]:
        """file id / آیدی فایل"""
        return self._data['file']['file_id'] if "file" in self._data else None
    @property
    def file_name(self) -> Optional[str]:
        """file name / اسم فایل"""
        return self._data['file']['file_name'] if "file" in self._data else None
    @property
    def size_file(self) -> Optional[int]:
        """size file / سایز فایل"""
        return self._data['file']['size'] if "file" in self._data else None
    @property
    def type_file(self) -> str:
        """get type file / گرفتن نوع فایل"""
        if self.file_name:
            return get_file_category(self.file_name)
        else:
            return "text"
    
    # Button

    @property
    def button(self) -> Optional[dict]:
        """data button clicked / اطلاعات دکمه کلیک شده"""
        return self._data['aux_data'] if "aux_data" in self._data else None
    @property
    def button_id(self) -> Optional[str]:
        """button id clicked button / آیدی دکمه کلیک شده"""
        return self.button['button_id'] if self.button else None
    
    # Reply

    @property
    def is_reply(self) -> bool:
        """is replyed / ریپلای شده"""
        return self.reply_to_message_id != None

    @property
    def reply_to_message_id(self) -> Optional[str]:
        """message id replyed / آیدی پیام ریپلای شده"""
        return self._data["reply_to_message_id"] if "reply_to_message_id" in self._data else None
    
    # Mata Data

    @property
    def metadata(self) -> Optional[dict]:
        """meta data / متا دیتا"""
        return self._data["metadata"] if "metadata" in self._data else None
    @property
    def meta_data_parts(self) -> Optional[metadata_prop]:
        """meta data parts list / لیست قسمت های متا دیتا"""
        return metadata_prop(self.metadata["meta_data_parts"]) if self.metadata else None
    
    # Forward

    @property
    def is_fowrard(self) -> bool:
        """forwarded / فوروارد شده"""
        return "forwarded_from" in self._data
    @property
    def forward_from(self) -> Optional[str]:
        """forward from / فوروارد از"""
        if self.is_fowrard:
            return self._data["forwarded_from"]["type_from"]
        return None
    @property
    def forward_message_id(self) -> Optional[str]:
        """message id forward / آیدی پیام فوروارد شده"""
        if self.is_fowrard:
            return self._data["forwarded_from"]["message_id"]
        return None
    @property
    def forward_from_sender_id(self) -> Optional[str]:
        """sender id forwarded / شناسه گوید فوروارد کننده"""
        if self.is_fowrard:
            return self._data["forwarded_from"]["from_sender_id"]
        return None
    
    # Contact

    @property
    def is_contact(self) -> bool:
        """contect / مخاطب"""
        return "contact_message" in self._data
    @property
    def contact_phone_number(self) -> Optional[str]:
        """phone number contact / شماره همراه مخاطب"""
        if self.is_contact:
            return self._data["contact_message"]["phone_number"]
        return None
    @property
    def contact_first_name(self) -> Optional[str]:
        """first name contact / اسم مخاطب"""
        if self.is_contact:
            return self._data["contact_message"]["first_name"]
        return None
    @property
    def contact_last_name(self) -> Optional[str]:
        """last name contact / نام خانوادگی مخاطب"""
        if self.is_contact:
            return self._data["contact_message"]["last_name"]
        return None

    # Stiker

    @property
    def is_sticker(self) -> bool:
        """sticker / استیکر"""
        return "sticker" in self._data
    @property
    def sticker_emoji_character(self) -> Optional[str]:
        """imoji sticker character / کاراکتر ایموجی استیکر"""
        if self.is_sticker:
            return self._data["sticker"]["emoji_character"]
        return None
    @property
    def sticker_sticker_id(self) -> Optional[str]:
        """sticker id / آیدی استیکر"""
        if self.is_sticker:
            return self._data["sticker"]["sticker_id"]
        return None
    @property
    def sticker_file(self) -> Union[File, dict]:
        """file sticker / فایل استیکر"""
        if self.is_sticker:
            return File(self._data["sticker"]["file"])
        return {}

    # remove update
    @property
    def removed_message_id(self) -> Optional[str]:
        """message id deleted / آیدی پیام پاک شده"""
        return self.raw_data_.get("removed_message_id")
    
    @property
    def update_time(self) -> Optional[int]:
        """update time message deleted / زمان پاک شدن پیام"""
        up_time = self.raw_data_.get("update_time")
        if up_time:
            return int(up_time)
        return None



    def regex(
        self,
        pattern: str,
        flags: int = 0,
        text: Optional[str] = None
    ) -> bool:
        """بررسی با الگو ریجکس / checking with regex"""
        if self.text is None:
            return False
        self.pattern = re.compile(pattern, flags)
        return bool(
            self.pattern.search(
                Utils.prefer_first(text, self.text)
            )
        )

    @auto_async
    async def get_chat_id_info(
        self,
        chat_id: Optional[str] = None
    ) -> Chat:
        """get info the chat id / گرفتن درباره چت آیدی"""
        return await self._client.get_chat(
            chat_id=Utils.prefer_first(chat_id, self.chat_id)
        )

    @auto_async
    async def reply(
        self,
        text: Optional[str] = None,
        keypad_inline: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: bool | None = True,
        on_time_keyboard: bool | None = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        # file
        file: Union[str , Path , bytes , None] = None,
        name_file: Optional[str] = None,
        type_file: Literal["File", "Image", "Voice", "Music", "Gif" , "Video"] = "File",
        file_id: Optional[str] = None,
        show_progress: bool = True,
        chunk_size: int = 64 * 1024,
        # poll
        question: Optional[str] = None,
        options: Optional[list] = None,
        type_poll: Literal["Regular", "Quiz"] = "Regular",
        is_anonymous: bool = True,
        correct_option_index: Optional[int] = None,
        allows_multiple_answers: bool = False,
        hint: Optional[str] = None,
        # location
        latitude: Optional[str] = None,
        longitude: Optional[str] = None,
        # contact
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        phone_number: Optional[str] = None,

        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None
    ) -> 'msg_update':
        """reply message / ریپلای پیام"""
        return await self._client.send_message(
            text=text,
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            inline_keypad=inline_keypad if inline_keypad else keypad_inline,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            keypad=keypad,
            on_time_keyboard=on_time_keyboard,
            resize_keyboard=resize_keyboard,
            meta_data=meta_data,
            file=file,
            name_file=name_file,
            type_file=type_file,
            file_id=file_id,
            show_progress=show_progress,
            question=question,
            options=options,
            type_poll=type_poll,
            is_anonymous=is_anonymous,
            correct_option_index=correct_option_index,
            allows_multiple_answers=allows_multiple_answers,
            hint=hint,
            latitude=latitude,
            longitude=longitude,
            first_name=first_name,
            last_name=last_name,
            phone_number=phone_number,
            chunk_size=chunk_size
        )

    @auto_async
    async def reply_text(
        self,
        text: str,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: bool | None = True,
        on_time_keyboard: bool | None = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None
    ) -> 'msg_update':
        """reply text / ریپلای متن"""
        return await self._client.send_text(
            text=text,
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            inline_keypad=inline_keypad,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            keypad=keypad,
            on_time_keyboard=on_time_keyboard,
            resize_keyboard=resize_keyboard,
            meta_data=meta_data
        )

    @auto_async
    async def reply_poll(
        self,
        question: str,
        options: list,
        type_poll: Literal["Regular", "Quiz"] = "Regular",
        is_anonymous: bool = True,
        correct_option_index: Optional[int] = None,
        allows_multiple_answers: bool = False,
        hint: Optional[str] = None,
        auto_delete: Optional[int] = None,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None
    ) -> 'msg_update':
        """reply poll / ریپلای نظرسنجی"""
        return await self._client.send_poll(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            question=question,
            options=options,
            type_poll=type_poll,
            is_anonymous=is_anonymous,
            correct_option_index=correct_option_index,
            allows_multiple_answers=allows_multiple_answers,
            hint=hint,
            auto_delete=auto_delete,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id)
        )

    @auto_async
    async def reply_contact(
        self,
        first_name: str,
        phone_number: str,
        last_name: str = "",
        auto_delete: Optional[int] = None,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None
    ) -> 'msg_update':
        """reply contact / ریپلای مخاطب"""
        return await self._client.send_contact(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            first_name=first_name,
            last_name=last_name,
            phone_number=phone_number,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            auto_delete=auto_delete
        )

    @auto_async
    async def reply_location(
        self,
        latitude: str,
        longitude: str,
        auto_delete: Optional[int] = None,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None
    ) -> 'msg_update':
        """reply location / ریپلای موقعیت مکانی"""
        return await self._client.send_location(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            latitude=latitude,
            longitude=longitude,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            auto_delete=auto_delete
        )

    @auto_async
    async def reply_file(
        self,
        file: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        type_file: Literal["File", "Image", "Voice", "Music", "Gif","Video"] = "File",
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply file / ریپلای فایل"""
        return await self._client.base_send_file(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            file=file,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            type_file=type_file,
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )
    
    @auto_async
    async def reply_document(
        self,
        file: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        type_file: Literal["File", "Image", "Voice", "Music", "Gif","Video"] = "File",
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply file / ریپلای فایل"""
        return await self._client.base_send_file(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            file=file,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            type_file=type_file,
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def reply_image(
        self,
        image: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply image / رپیلای تصویر"""
        return await self._client.send_image(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            image=image,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def reply_voice(
        self,
        voice: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply voice / رپیلای ویس"""
        return await self._client.send_voice(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            voice=voice,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def reply_music(
        self,
        music: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply voice / رپیلای موزیک"""
        return await self._client.send_music(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            music=music,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def reply_gif(
        self,
        gif: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply voice / رپیلای گیف"""
        return await self._client.send_gif(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            gif=gif,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def reply_video(
        self,
        video: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        disable_notification: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        reply_to_message_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ) -> 'msg_update':
        """reply voice / رپیلای ویدیو"""
        return await self._client.send_video(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            video=video,
            name_file=name_file,
            text=text,
            reply_to_message_id=Utils.prefer_first(reply_to_message_id, self.message_id),
            disable_notification=disable_notification,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def forward(
        self,
        to_chat_id: str,
        auto_delete: Optional[int] = None,
        from_chat_id: Optional[str] = None,
        message_id: Optional[str] = None
    ) -> 'msg_update':
        """forward / فوروارد"""
        return await self._client.forward_message(
            from_chat_id=Utils.prefer_first(from_chat_id, self.chat_id),
            message_id=Utils.prefer_first(message_id, self.message_id),
            to_chat_id=to_chat_id,
            auto_delete=auto_delete
        )

    @auto_async
    async def download(
        self,
        path : str = "file",
        file_id: Optional[str] = None,
        show_progress: bool = True
    ) -> None:
        """download / دانلود"""
        final_id = file_id or self.file_id
        if final_id is None:
            raise TypeError("Message is not file and you not got the file_id Arg.")
        await self._client.download_file(
            id_file=final_id,
            path=path,
            show_progress=show_progress
        )
    
    @auto_async
    async def get_download_file_url(
        self,
        file_id: Optional[str] = None
    ) -> Optional[str]:
        """getting url download file / گرفتن لینک دانلود فایل"""
        final_id = file_id or self.file_id
        if final_id is None:
            raise TypeError("Message is not file and you not got the file_id Arg.")
        return await self._client.get_download_file_url(
            id_file=final_id
        )

    @auto_async
    async def delete(
        self,
        chat_id: Optional[str] = None,
        message_id: Optional[str] = None
    ) -> props:
        """delete / حذف"""
        return await self._client.delete_message(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            message_id=Utils.prefer_first(message_id, self.message_id)
        )
    
    @auto_async
    async def ban(
        self,
        chat_id: Optional[str] = None,
        user_id: Optional[str] = None
    ):
        """ban user / بن کاربر"""
        return await self._client.ban_chat_member(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            user_id=Utils.prefer_first(user_id, self.sender_id),
        )
    
    @auto_async
    async def unban(
        self,
        chat_id: Optional[str] = None,
        user_id: Optional[str] = None
    ):
        """un ban user / آنبن کاربر"""
        return await self._client.unban_chat_member(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            user_id=Utils.prefer_first(user_id, self.sender_id),
        )

    @auto_async
    async def ban_reply(
        self,
        chat_id: Optional[str] = None,
    ):
        if not self.reply_to_message_id:
            return None
        msg = await self._client.get_message(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            message_id=self.reply_to_message_id
        )
        if not msg:
            return None
        return await self._client.ban_chat_member(
            chat_id=self.chat_id,
            user_id=msg.sender_id
        )

    @auto_async
    async def unban_reply(
        self,
        chat_id: Optional[str] = None
    ):
        if not self.reply_to_message_id:
            return None
        msg = await self._client.get_message(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            message_id=self.reply_to_message_id
        )
        if not msg:
            return None
        return await self._client.unban_chat_member(
            chat_id=self.chat_id,
            user_id=msg.sender_id
        )

    @auto_async
    async def get_reply(
        self,
        chat_id: Optional[str] = None,
        message_id: Optional[str] = None
    ):
        """get message replyed / گرفتن پیام ریپلای شده"""
        if not self.is_reply:
            return None
        msg = await self._client.get_message(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            message_id=Utils.prefer_first(message_id, self.reply_to_message_id)
        )
        return msg
    

    def to_dict(
        self
    ) -> dict:
        """convert Update to Dict / تبدیل آپدیت به دیکشنری"""
        if isinstance(self._data, props):
            self._data = dict(self._data.to_dict())
        if self.file_name:
            self._data['file']['type'] = self.type_file
        self._data["sender_type"] = self.sender_type
        return self._data

    def __str__(self) -> str:
        return json.dumps(
            self.to_dict(),
            indent=4,
            ensure_ascii=False
        )

    def __repr__(self) -> str:
        return self.__str__()


class UpdateButton:
    def __init__(self, data: dict, client: "Client"):
        self._data = data
        self._client = client

    @property
    def raw_data(self) -> dict:
        return self._data

    @property
    def button_id(self) -> str:
        """button id clicked / آیدی دکمه کلیک شده"""
        return self._data["inline_message"]["aux_data"]["button_id"]

    @property
    def chat_id(self) -> str:
        """chat id clicked / چت آیدی کلیک شده"""
        return self._data["inline_message"]["chat_id"]

    @property
    def message_id(self) -> str:
        """message id for message clicked glass button / آیدی پیام کلیک شده روی دکمه شیشه ای"""
        return self._data["inline_message"]["message_id"]

    @property
    def sender_id(self) -> str:
        """guid for clicked button glass / شناسه گوید کاربر کلیک کرده روی دکمه شیشه ای"""
        return self._data["inline_message"]["sender_id"]

    @property
    def text(self) -> str:
        """text for button clicked / متن دکمه شیشه ای که روی آن کلیک شده"""
        return self._data["inline_message"]["text"]

    @auto_async
    async def send_text(
        self,
        text: str,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML'] = "Markdown",
        meta_data: Optional[list] = None,
        chat_id: Optional[str] = None
    ):
        """send text / ارسال متن"""
        return await self._client.send_text(
            text=text,
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            inline_keypad=inline_keypad,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            meta_data=meta_data
        )
    
    @auto_async
    async def send_message(
        self,
        text: Optional[str] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        reply_to_message_id: Optional[str] = None,
        auto_delete: Optional[int] = None,
        parse_mode: Literal["Markdown","HTML",None] = "Markdown",
        meta_data: Optional[list] = None,
        # file
        file: Union[str , Path , bytes , None] = None,
        name_file: Optional[str] = None,
        type_file: Literal["File", "Image", "Voice", "Music", "Gif" , "Video"] = "File",
        file_id: Optional[str] = None,
        show_progress: bool = True,
        chunk_size: int = 64 * 1024,
        # poll
        question: Optional[str] = None,
        options: Optional[list] = None,
        type_poll: Literal["Regular", "Quiz"] = "Regular",
        is_anonymous: bool = True,
        correct_option_index: Optional[int] = None,
        allows_multiple_answers: bool = False,
        hint: Optional[str] = None,
        # location
        latitude: Optional[str] = None,
        longitude: Optional[str] = None,
        # contact
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        phone_number: Optional[str] = None,

        chat_id: Optional[str] = None
    ):
        """send message / ارسال پیام"""
        return await self._client.send_message(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            text=text,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            reply_to_message_id=reply_to_message_id,
            auto_delete=auto_delete,
            parse_mode=parse_mode,
            meta_data=meta_data,
            file=file,
            name_file=name_file,
            type_file=type_file,
            file_id=file_id,
            show_progress=show_progress,
            question=question,
            options=options,
            type_poll=type_poll,
            is_anonymous=is_anonymous,
            correct_option_index=correct_option_index,
            allows_multiple_answers=allows_multiple_answers,
            hint=hint,
            latitude=latitude,
            longitude=longitude,
            first_name=first_name,
            last_name=last_name,
            phone_number=phone_number,
            chunk_size=chunk_size
        )

    @auto_async
    async def send_pool(
        self,
        question: str,
        options : list,
        type_poll: Literal['Regular', 'Quiz'] = "Regular",
        is_anonymous: bool = True,
        correct_option_index: Optional[int] = None,
        allows_multiple_answers: bool = False,
        hint: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        chat_id: Optional[str] = None
    ):
        """sending pool / ارسال نظرسنجی"""
        return await self._client.send_poll(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            question=question,
            options=options,
            type_poll=type_poll,
            is_anonymous=is_anonymous,
            correct_option_index=correct_option_index,
            allows_multiple_answers=allows_multiple_answers,
            hint=hint,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id
        )

    @auto_async
    async def send_contact(
        self,
        first_name: str,
        last_name: str,
        phone_number: str,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        inline_keypad: Optional[list] = None,
        chat_id: Optional[str] = None
    ):
        """sending contact / ارسال مخاطب"""
        return await self._client.send_contact(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            first_name=first_name,
            last_name=last_name,
            phone_number=phone_number,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            inline_keypad=inline_keypad
        )

    @auto_async
    async def send_location(
        self,
        latitude: str,
        longitude: str,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        chat_id: Optional[str] = None
    ):
        """sending location / ارسال موقعیت مکانی"""
        return await self._client.send_location(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            latitude=latitude,
            longitude=longitude,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id
        )

    @auto_async
    async def send_file(
        self,
        file: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        type_file: Literal['File', 'Image', 'Voice', 'Music', 'Gif', 'Video'] = "File",
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending file / ارسال فایل"""
        return await self._client.base_send_file(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            file=file,
            name_file=name_file,
            text=text,
            type_file=type_file,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )
    
    @auto_async
    async def send_document(
        self,
        file: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending file / ارسال فایل"""
        return await self._client.send_document(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            file=file,
            name_file=name_file,
            text=text,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def send_image(
        self,
        image: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending image / ارسال تصویر"""
        return await self._client.send_image(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            image=image,
            name_file=name_file,
            text=text,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def send_video(
        self,
        video: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending video / ارسال ویدیو"""
        return await self._client.send_video(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            video=video,
            name_file=name_file,
            text=text,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def send_voice(
        self,
        voice: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending voice / ارسال ویس"""
        return await self._client.send_voice(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            voice=voice,
            name_file=name_file,
            text=text,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def send_music(
        self,
        music: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending music / ارسال موزیک"""
        return await self._client.send_music(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            music=music,
            name_file=name_file,
            text=text,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )

    @auto_async
    async def send_gif(
        self,
        gif: Union[str , Path , bytes],
        name_file: Optional[str] = None,
        text: Optional[str] = None,
        auto_delete: Optional[int] = None,
        reply_to_message_id: Optional[str] = None,
        parse_mode: Literal['Markdown', 'HTML', None] = "Markdown",
        meta_data: Optional[list] = None,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        upload_by: Literal["aiohttp", "httpx"] = "aiohttp",
        show_progress: bool = True,
        chat_id: Optional[str] = None,
        chunk_size: int = 64 * 1024
    ):
        """sending gif / ارسال گیف"""
        return await self._client.send_gif(
            chat_id=Utils.prefer_first(chat_id, self.chat_id),
            gif=gif,
            name_file=name_file,
            text=text,
            auto_delete=auto_delete,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            meta_data=meta_data,
            inline_keypad=inline_keypad,
            keypad=keypad,
            resize_keyboard=resize_keyboard,
            on_time_keyboard=on_time_keyboard,
            upload_by=upload_by,
            show_progress=show_progress,
            chunk_size=chunk_size
        )
    
    def to_dict(
        self
    ) -> dict:
        """convert Update to Dict / تبدیل آپدیت به دیکشنری"""
        if isinstance(self._data, props):
            self._data = dict(self._data.to_dict())
        return self._data

    def __str__(self) -> str:
        return json.dumps(
            self.to_dict(),
            indent=4,
            ensure_ascii=False
        )

    def __repr__(self) -> str:
        return self.__str__()




from json import dumps, loads
from tqdm import tqdm
from typing import Any, Protocol, Optional, Union
import httpx
import aiohttp
from ..utils import Configs
from ..exceptions import *
from .helper import Helper
from ..utils import *


class MethodsProtocol(Protocol):
    sessionData: Any
    crypto: Any
    proxy: Optional[str]
    platform: str
    apiVersion: int
    timeOut: int
    showProgressBar: bool
    async def requestSendFile(
        self, fileName: str, mime: str, size: int
    ) -> dict: ...


class Network:
    def __init__(self, methods: MethodsProtocol) -> None:
        self.methods = methods
        self.sessionData = methods.sessionData
        self.crypto = methods.crypto
        
        if methods.proxy:
            import os
            if isinstance(methods.proxy, dict):
                proxy_str = methods.proxy.get("http") or methods.proxy.get("https") or list(methods.proxy.values())[0]
            else:
                proxy_str = methods.proxy
            
            os.environ['HTTP_PROXY'] = proxy_str
            os.environ['HTTPS_PROXY'] = proxy_str
            os.environ['ALL_PROXY'] = proxy_str
        
        self.httpx_client = httpx.AsyncClient(
            timeout=methods.timeOut,
            trust_env=True
        )
        
        self.proxy_str = None
        if methods.proxy:
            if isinstance(methods.proxy, dict):
                self.proxy_str = methods.proxy.get("http") or methods.proxy.get("https") or list(methods.proxy.values())[0]
            else:
                self.proxy_str = methods.proxy
        
        self.aiohttp_session = None

    async def _get_aiohttp_session(self):
        if self.aiohttp_session is None:
            connector = aiohttp.TCPConnector()
            
            if self.proxy_str:
                self.aiohttp_session = aiohttp.ClientSession(
                    connector=connector,
                    trust_env=True
                )
            else:
                self.aiohttp_session = aiohttp.ClientSession(
                    connector=connector
                )
        
        return self.aiohttp_session

    async def request(self, method: str, input: dict = {}, tmpSession: bool = False, 
                     attempt: int = 0, maxAttempt: int = 2) -> dict:
        url: str = Helper.getApiServer()
        platform: str = self.methods.platform.lower()
        apiVersion: int = self.methods.apiVersion
        configs = Configs()
        
        if platform in ["rubx", "rubikax"]:
            client: dict = configs.clients["android"]
            client["package"] = "ir.rubx.bapp"
        elif platform in ["android"]:
            client: dict = configs.clients["android"]
        else:
            client: dict = configs.clients["web"]

        auth_key = "tmp_session" if tmpSession else "auth"
        auth_value = (
            self.crypto.auth if tmpSession else
            self.crypto.changeAuthType(self.sessionData["auth"]) if apiVersion > 5 else
            self.sessionData["auth"]
        )
        
        data = {
            "api_version": str(apiVersion),
            auth_key: auth_value,
            "data_enc": self.crypto.encrypt(
                dumps({
                    "method": method,
                    "input": input,
                    "client": client
                })
            )
        }

        headers: dict = {
            "Referer": "https://web.rubika.ir/",
            "Content-Type": "application/json; charset=utf-8",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
        }

        if not tmpSession and apiVersion > 5:
            data["sign"] = self.crypto.makeSignFromData(data["data_enc"])

        while attempt <= maxAttempt:
            try:
                response = await self.httpx_client.post(
                    url=url,
                    headers=headers,
                    json=data
                )
                response.raise_for_status()
                
                result_data = response.json()
                decrypted_data = self.crypto.decrypt(result_data["data_enc"])
                result = loads(decrypted_data)
                
                if result["status"] == "OK":
                    if tmpSession:
                        result["data"]["tmp_session"] = self.crypto.auth
                    return result["data"]
                else:
                    raise Utils.raise_error(result)
                    
            except (httpx.RequestError, httpx.HTTPStatusError, ValueError) as e:
                attempt += 1
                if attempt > maxAttempt:
                    raise httpx.NetworkError(f"Request failed after {maxAttempt} attempts: {str(e)}")
                continue

        raise httpx.NetworkError("Request failed")

    async def upload(
        self, 
        file: Union[str, bytes], 
        fileName: Optional[str] = None, 
        chunkSize: int = 131072
    ) -> Optional[dict]:
        if isinstance(file, str):
            if Utils.checkLink(url=file):
                response = await self.httpx_client.get(file)
                response.raise_for_status()
                file_bytes: bytes = response.content
                mime: str = Utils.getMimeFromByte(file_bytes)
                fileName = fileName or Utils.generateFileName(mime=mime)
                file = file_bytes
            else:
                fileName = fileName or file
                with open(file, "rb") as fh:
                    file = fh.read()
                mime = Utils.getMimeFromByte(file)

        elif isinstance(file, bytes):
            mime = Utils.getMimeFromByte(file)
            fileName = fileName or Utils.generateFileName(mime=mime)
        else:
            raise FileNotFoundError("Enter a valid path or url or bytes of file.")

        async def send_chunk(session: aiohttp.ClientSession, data: bytes, 
                           upload_url: str, headers: dict, maxAttempts: int = 2) -> Optional[dict]:
            for attempt in range(maxAttempts):
                try:
                    async with session.post(
                        upload_url,
                        headers=headers,
                        data=data,
                        timeout=aiohttp.ClientTimeout(total=self.methods.timeOut)
                    ) as response:
                        response.raise_for_status()
                        return await response.json()
                except Exception as e:
                    print(f"\nError uploading file! (Attempt {attempt + 1}/{maxAttempts}): {e}")
            
            print("\nFailed to upload the file!")
            return None

        requestSendFileData = await self.methods.requestSendFile(
            fileName=fileName,
            mime=mime,
            size=len(file)
        )

        session = await self._get_aiohttp_session()
        
        headers = {
            "auth": self.sessionData["auth"],
            "access-hash-send": requestSendFileData["access_hash_send"],
            "file-id": requestSendFileData["id"],
        }

        totalParts = (len(file) + chunkSize - 1) // chunkSize

        processBar: Optional[tqdm] = None

        if self.methods.showProgressBar:
            processBar = tqdm(
                desc=f"Uploading {fileName}",
                total=len(file),
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        for partNumber in range(1, totalParts + 1):
            startIdx = (partNumber - 1) * chunkSize
            endIdx = min(startIdx + chunkSize, len(file))
            chunk_headers = headers.copy()
            chunk_headers["chunk-size"] = str(endIdx - startIdx)
            chunk_headers["part-number"] = str(partNumber)
            chunk_headers["total-part"] = str(totalParts)
            
            data = file[startIdx:endIdx]
            hashFileReceive = await send_chunk(
                session, 
                data, 
                requestSendFileData["upload_url"],
                chunk_headers
            )
            
            if processBar is not None:
                processBar.update(len(data))

            if not hashFileReceive:
                return None
            
            if partNumber == totalParts:
                if not hashFileReceive.get("data"):
                    return None
                
                requestSendFileData["file"] = file
                requestSendFileData["access_hash_rec"] = hashFileReceive["data"]["access_hash_rec"]
                requestSendFileData["file_name"] = fileName
                requestSendFileData["mime"] = mime
                requestSendFileData["size"] = len(file)
                return requestSendFileData
        
        return None

    async def download(self, accessHashRec: str, fileId: str, dcId: str, 
                      size: int, fileName: str, chunkSize: int = 262143, 
                      attempt: int = 0, maxAttempts: int = 2) -> Optional[bytes]:
        headers = {
            "auth": self.sessionData["auth"],
            "access-hash-rec": accessHashRec,
            "dc-id": dcId,
            "file-id": fileId,
            "Host": f"messenger{dcId}.iranlms.ir",
            "client-app-name": "Main",
            "client-app-version": "3.5.7",
            "client-package": "app.rbmain.a",
            "client-platform": "Android",
            "Connection": "Keep-Alive",
            "Content-Type": "application/json",
            "User-Agent": "okhttp/3.12.1"
        }

        session = await self._get_aiohttp_session()
        url = f"https://messenger{dcId}.iranlms.ir/GetFile.ashx"

        processBar: Optional[tqdm] = None

        if self.methods.showProgressBar:
            processBar = tqdm(
                desc=f"Downloading {fileName}",
                total=size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        for retry in range(maxAttempts):
            try:
                data: bytes = b""
                
                async with session.post(
                    url,
                    headers=headers,
                    timeout=aiohttp.ClientTimeout(total=self.methods.timeOut)
                ) as response:
                    response.raise_for_status()
                    
                    async for chunk in response.content.iter_chunked(chunkSize):
                        if chunk:
                            data += chunk
                            if processBar is not None:
                                processBar.update(len(chunk))
                        
                        if len(data) >= size:
                            if processBar is not None:
                                processBar.close()
                            return data[:size]
                
            except Exception as e:
                attempt += 1
                if attempt <= maxAttempts:
                    print(f"\nError downloading file! (Attempt {attempt}/{maxAttempts}): {e}")
                    continue
                
                if processBar is not None:
                    processBar.close()
                raise TimeoutError("Failed to download the file!")
        
        if processBar is not None:
            processBar.close()
        return None

    async def close(self):
        """Close all HTTP sessions"""
        await self.httpx_client.aclose()
        
        if self.aiohttp_session:
            await self.aiohttp_session.close()
            self.aiohttp_session = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



from enum import Enum

class ParseMode(str, Enum):
    MARKDOWN = "Markdown"
    HTML = "HTML"

class Style(str, Enum):
    BOLD = "bold"
    SPOIL = "spoil"
    UNDERLINE = "underline"
    STRIKE = "strike"
    COPY = "copy"
    CODE = "code"
    LINK = "link"
    MENTION = "mention"

class MetaData:
    """متا دادن متن به صورت بولد، اسپویل، آندرلاین و ..."""
    
    _TEMPLATES = {
        Style.BOLD: {ParseMode.MARKDOWN: "**#text**", ParseMode.HTML: "<b>#text</b>"},
        Style.SPOIL: {ParseMode.MARKDOWN: "||#text||", ParseMode.HTML: '<span class="tg-spoiler">#text</span>'},
        Style.UNDERLINE: {ParseMode.MARKDOWN: "__#text__", ParseMode.HTML: "<u>#text</u>"},
        Style.STRIKE: {ParseMode.MARKDOWN: "~~#text~~", ParseMode.HTML: "<s>#text</s>"},
        Style.COPY: {ParseMode.MARKDOWN: "`#text`", ParseMode.HTML: "<code>#text</code>"},
        Style.CODE: {ParseMode.MARKDOWN: "```#text```", ParseMode.HTML: "<pre>#text</pre>"},
        Style.LINK: {ParseMode.MARKDOWN: "[#text](#link)", ParseMode.HTML: '<a href="#link">#text</a>'},
        Style.MENTION: {ParseMode.MARKDOWN: "[#text](#sender_id)", ParseMode.HTML: '<mention objectId="#sender_id">#text</mention>'}
    }

    def __init__(self, parse_mode: ParseMode = ParseMode.MARKDOWN):
        self.parse_mode = parse_mode

    def format_text(self, text: str, style: Style, **kwargs) -> str:
        """متن را با استایل مشخص قالب‌بندی می‌کند"""
        try:
            template = self._TEMPLATES[style][self.parse_mode]
        except KeyError:
            raise ValueError(f"Unsupported style or parse mode: {style}, {self.parse_mode}")
        for k, v in kwargs.items():
            template = template.replace(f"#{k}", v)
        return template.replace("#text", text)

    def bold(self, text: str) -> str:
        """bold text / متن بولد"""
        return self.format_text(text, Style.BOLD)

    def spoil(self, text: str) -> str:
        """spoil text / متن اسپویل"""
        return self.format_text(text, Style.SPOIL)

    def underline(self, text: str) -> str:
        """underline text / متن آندرلاین"""
        return self.format_text(text, Style.UNDERLINE)

    def strike(self, text: str) -> str:
        """strike text / متن خط خورده"""
        return self.format_text(text, Style.STRIKE)

    def copy(self, text: str) -> str:
        """copy text / متن کپی"""
        return self.format_text(text, Style.COPY)

    def code(self, text: str) -> str:
        """code text / متن کد"""
        return self.format_text(text, Style.CODE)

    def link(self, text: str, link: str) -> str:
        """link text / متن لینک"""
        return self.format_text(text, Style.LINK, link=link)

    def mention(self, text: str, sender_id: str) -> str:
        """mention text / متن منشن"""
        return self.format_text(text, Style.MENTION, sender_id=sender_id)



from .core.client import Client
from .utils import filters
from .utils.version import v

__author__ = "Seyyed Mohamad Hosein Moosavi Raja"
__version__ = v
__all__ = ['Client', 'filters', '__version__']


from typing import Optional, Union, Dict, Any, List, Literal
from tqdm.asyncio import tqdm
from pathlib import Path
from ..type.errors import ServerRubikaError
import os
import aiofiles
import asyncio
import httpx
import tempfile
import logging


class Network:
    def __init__(
        self,
        token: str,
        logger: Optional[logging.Logger] = None,
        max_retries: int = 3,
        user_agent: Optional[str] = None,
        main_url: str = "https://botapi.rubika.ir/v3/",
        proxy: Optional[str] = None,
        rate_limit: int = 20
    ):
        self.logger = logger or logging.getLogger("fast_rub.network")
        self.token = token
        self.user_agent = user_agent
        self.main_url = main_url
        self.proxy = proxy
        self.max_retries = max_retries

        self._client: httpx.AsyncClient | None = None
        self._client_lock = asyncio.Lock()

        self._rate_sem: Optional[asyncio.Semaphore] = None
        self._queue: Optional[asyncio.Queue] = None
        self._worker_task: Optional[asyncio.Task] = None
        self._closed = False
        self._rate_limit = rate_limit

    # -------------------
    # Lifecycle: start / stop
    # -------------------
    async def start(self, *, start_worker: bool = True):
        asyncio.get_running_loop()

        if self._client_lock is None:
            self._client_lock = asyncio.Lock()
        if self._rate_sem is None:
            self._rate_sem = asyncio.Semaphore(self._rate_limit)
        if self._queue is None:
            self._queue = asyncio.Queue()

        await self._create_client()

        if start_worker:
            await self.start_worker()

    async def _create_client(self):
        self._client = httpx.AsyncClient(**self._build_client_kwargs())
        self.logger.debug("HTTP client created")

    async def close(self):
        self._closed = True

        if self._worker_task:
            self._worker_task.cancel()
            try:
                await self._worker_task
            except asyncio.CancelledError:
                pass

        if self._client and not self._client.is_closed:
            try:
                await self._client.aclose()
            except RuntimeError as e:
                if "Event loop is closed" not in str(e):
                    raise


    # -------------------
    # Worker queue
    # -------------------
    async def start_worker(self):
        if self._worker_task is None or self._worker_task.done():
            self._worker_task = asyncio.create_task(self._worker())
            self.logger.debug("Worker started")

    def _build_client_kwargs(self) -> dict:
        kwargs = {
            "timeout": httpx.Timeout(connect=10.0, read=30.0, write=10.0, pool=10.0),
            "limits": httpx.Limits(max_connections=100, max_keepalive_connections=20),
            "http1": True,
            "http2": True,
            "headers": {
                "Content-Type": "application/json",
                "User-Agent": self.user_agent or "fast_rub/Network"
            }
        }

        if self.proxy:
            try:
                httpx.AsyncClient(proxy=self.proxy)
                kwargs["proxies"] = self.proxy
            except TypeError:
                kwargs["proxy"] = self.proxy

        return kwargs


    async def _worker(self):
        if self._queue is None:
            self.logger.warning("Worker started but queue is None.")
            return

        while not self._closed:
            try:
                url, method, data_, headers, overrides, fut = await self._queue.get()
            except asyncio.CancelledError:
                break
            try:
                res = await self._do_request(url, method, data_, headers, **overrides)
                if not fut.done():
                    fut.set_result(res)
            except Exception as e:
                if not fut.done():
                    fut.set_exception(e)
            finally:
                try:
                    self._queue.task_done()
                except Exception:
                    pass

    # -------------------
    # Public request
    # -------------------
    async def request(
        self,
        url: str,
        data_: Optional[Union[Dict[str, Any], List[Any]]] = None,
        type_send: Literal["POST", "GET"] = "POST",
        *,
        max_retries: Optional[int] = None,
        timeout: Optional[float] = None
    ) -> httpx.Response:
        if self._closed:
            raise RuntimeError("Network is closed")

        overrides = {
            "max_retries": max_retries or self.max_retries,
            "timeout": timeout or 30.0
        }

        return await self._do_request(url, type_send, data_, None, **overrides)

    # -------------------
    # Internal request with retry
    # -------------------
    async def _ensure_client(self):
        if self._client_lock is None:
            self._client_lock = asyncio.Lock()
        if self._rate_sem is None:
            self._rate_sem = asyncio.Semaphore(self._rate_limit)
        if self._queue is None:
            self._queue = asyncio.Queue()

        if self._client is None or (hasattr(self._client, "is_closed") and self._client.is_closed):
            async with self._client_lock:
                if self._client is None or (hasattr(self._client, "is_closed") and self._client.is_closed):
                    await self._create_client()

    async def _do_request(
        self,
        url: str,
        method: str,
        data_: Optional[Union[Dict, List]],
        headers: Optional[Dict[str, str]],
        *,
        max_retries: int,
        timeout: float
    ) -> httpx.Response:

        last_exception = None
        headers = headers.copy() if headers else {"Content-Type": "application/json"}
        if self.user_agent:
            headers["User-Agent"] = self.user_agent

        await self._ensure_client()

        for attempt in range(1, max_retries + 1):
            try:
                if self._rate_sem is None:
                    self._rate_sem = asyncio.Semaphore(self._rate_limit)

                async with self._rate_sem:
                    if method == "POST":
                        resp = await self._client.post(url, json=data_, headers=headers, timeout=timeout)
                    elif method == "GET":
                        resp = await self._client.get(url, headers=headers, timeout=timeout)
                    else:
                        raise ValueError(f"Invalid method: {method}")

                resp.raise_for_status()
                return resp

            except (httpx.ReadError, httpx.ConnectError, httpx.TimeoutException) as e:
                last_exception = e
                wait_time = attempt * 1.5
                self.logger.warning(f"Request failed attempt {attempt}/{max_retries}: {e}. Retrying in {wait_time}s...")

                if self._client_lock is None:
                    self._client_lock = asyncio.Lock()

                async with self._client_lock:
                    if self._client:
                        try:
                            await self._client.aclose()
                        except RuntimeError as re:
                            if "Event loop is closed" in str(re):
                                self.logger.debug("Event loop closed while closing client (during retry) — ignoring.")
                            else:
                                self.logger.exception("RuntimeError when closing client (during retry)")
                        except Exception:
                            pass
                        self._client = None

                try:
                    await self._ensure_client()
                except Exception as e2:
                    self.logger.exception(f"Failed to recreate client: {e2}")
                    raise e2

                await asyncio.sleep(wait_time)

            except Exception as e:
                self.logger.error(f"Unexpected error during request: {e}")
                raise e

        self.logger.error(f"All {max_retries} attempts failed for {url}")
        raise last_exception or RuntimeError("Unknown error in request")

    # -------------------
    # Rubika high-level
    # -------------------
    async def send_request(
        self,
        method: str,
        data_: Optional[Union[Dict[str, Any], List[Any]]] = None
    ) -> dict:
        self.logger.debug(f"method {method}")
        url = f"{self.main_url}{self.token}/{method}"
        resp = await self.request(url, data_, "POST")
        try:
            result = resp.json()
        except Exception:
            raise ServerRubikaError("Error converting response to JSON")

        from ..utils.utils import Utils
        if not Utils.check_data(result):
            self.logger.error(f"Server returned error: {result}")
            raise ServerRubikaError(result)
        return result["data"]

    async def download(self, url: str, path: str = "file", show_progress: bool = True) -> bool:
        try:
            await self._ensure_client()

            dir_path = os.path.dirname(path)
            if dir_path:
                os.makedirs(dir_path, exist_ok=True)

            if self._rate_sem is None:
                self._rate_sem = asyncio.Semaphore(self._rate_limit)

            async with self._rate_sem:
                async with self._client.stream("GET", url) as response:
                    response.raise_for_status()
                    total = int(response.headers.get("content-length", 0))
                    if show_progress and total:
                        pbar = tqdm(total=total, unit="B", unit_scale=True, desc="Downloading")
                    else:
                        pbar = None
                    async with aiofiles.open(path, 'wb') as file:
                        async for chunk in response.aiter_bytes():
                            await file.write(chunk)
                            if pbar:
                                pbar.update(len(chunk))
                    if pbar:
                        pbar.close()
            return True
        except Exception as e:
            self.logger.error(f"Download failed: {e}")
            return False
    
    async def upload_httpx(self, url: str, files: dict) -> dict:
        try:
            self.logger.info("در حال آپلود فایل ...")
            await self._ensure_client()

            from ..utils.utils import Utils
            response = await self._client.post(url, files=files, timeout=None)
            response.raise_for_status()
            result = response.json()
            if not Utils.check_data(result):
                self.logger.error(f"Server returned error: {result}")
                raise ServerRubikaError(result)
            self.logger.info("فایل آپلود شد")
            return result["data"]
        except httpx.HTTPError:
            self.logger.error("خطای HTTP")
            raise
        except Exception as e:
            self.logger.error(f"خطایی ناشناخته » {e}")
            raise

    async def upload(
        self,
        url: str,
        file_path: Union[str, Path, bytes],
        file_name: str,
        show_progress: bool = True,
        chunk_size: int = 64 * 1024
    ) -> dict:
        self.logger.info("در حال آپلود فایل (aiohttp stream)...")
        try:
            import aiohttp
        except ImportError:
            raise ImportError("ابتدا کتابخانه aiohttp را برای آپلود نصب کنید !")
        is_temp = False
        if isinstance(file_path, (bytes, bytearray)):
            tmp = tempfile.NamedTemporaryFile(delete=False)
            tmp.write(file_path)
            tmp.close()
            file_path = tmp.name
            is_temp = True
        try:
            total_size = None
            if isinstance(file_path, (str, Path)):
                total_size = os.path.getsize(file_path)
            timeout = aiohttp.ClientTimeout(total=None)
            async with aiohttp.ClientSession(timeout=timeout) as session:
                form = aiohttp.FormData()
                async def file_gen():
                    async with aiofiles.open(file_path, "rb") as f:
                        if show_progress and total_size:
                            pbar = tqdm(total=total_size, unit="B", unit_scale=True, desc="Uploading")
                        else:
                            pbar = None
                        while True:
                            chunk = await f.read(chunk_size)
                            if not chunk:
                                break
                            if pbar:
                                pbar.update(len(chunk))
                            yield chunk
                        if pbar:
                            pbar.close()
                form.add_field(
                    "file",
                    file_gen(),
                    filename=file_name,
                    content_type="application/octet-stream"
                )
                async with session.post(url, data=form) as resp:
                    if resp.status != 200:
                        text = await resp.text()
                        raise ServerRubikaError(
                            {"status": "ERROR", "detail": text}
                        )
                    result = await resp.json()
                    from ..utils.utils import Utils
                    if not Utils.check_data(result):
                        self.logger.error(f"Server returned error: {result}")
                        raise ServerRubikaError(result)
                    self.logger.info("آپلود با موفقیت انجام شد")
                    return result["data"]
        finally:
            if is_temp:
                os.remove(file_path)



from typing import Dict, Tuple, Optional, Union, TYPE_CHECKING
from pathlib import Path
import aiofiles
from .colors import cprint, Colors
import time
from .encryption import Encryption
import json
import os
from ..type.errors import CreateSessionError
if TYPE_CHECKING:
    from ..network.network import Network

class Utils:
    @staticmethod
    def format_file(type_file: Optional[str] = None) -> Optional[str]:
        if not type_file:
            return None
        for type_,pass_ in {"File":"", "Image":".png", "Voice":".mp3", "Music":".mp3", "Gif":".mp4" , "Video":".mp4"}.items():
            if type_ == type_file:
                name_file = type_+pass_
                return name_file
        return None
    
    @staticmethod
    def print_time(text: str, time_sleep: float = 0.07, color: str = Colors.WHITE) -> None:
        k = ""
        for text in text:
            k += text
            print(f"{color}{k}{Colors.RESET}", end="\r")
            time.sleep(time_sleep)
        cprint("",Colors.WHITE)
    
    @staticmethod
    def get_input(text_output: str) -> str:
        text = None
        while text is None or len(text) != 64:
            cprint("Write the valid ! Your text invalid.",Colors.RED)
            text = input(text_output)
        return text
    
    @staticmethod
    def calculate_upload_timeout(file_size_bytes: int, upload_speed_bps: int = 300_000) -> int:
        SAFETY_FACTOR = 1.5
        timeout_seconds = (file_size_bytes / upload_speed_bps) * SAFETY_FACTOR
        return max(int(timeout_seconds), 30)

    # Session
    
    @staticmethod
    def session_dict(name_session: str, token: str, user_agent: Optional[str] = None, time_out: Optional[float] = 30.0, display_welcome: bool = False, view_logs: Optional[bool] = False, save_logs: Optional[bool] = False) -> dict:
        text_json_fast_rub_session = {
            "name_session": name_session,
            "token": token,
            "user_agent": user_agent,
            "time_out": time_out,
            "display_welcome": display_welcome,
            "setting_logs":{
                "view":view_logs,
                "save":save_logs
            }
        }
        return text_json_fast_rub_session
    
    @staticmethod
    def save_dict(data: dict, name_file: str):
        json_fast_rub_session = json.dumps(data,indent=4,ensure_ascii=False)
        json_fast_rub_session = Encryption.en(str(json_fast_rub_session))
        Utils.save(json_fast_rub_session, name_file)

    @staticmethod
    def save(data: str, name_file: str):
        with open(name_file, "w", encoding="utf-8") as file:
            file.write(data)

    @staticmethod
    def create_session(name_session: str, token: Optional[str] = None, user_agent: Optional[str] = None, time_out: Optional[float] = 30.0, display_welcome: bool = False, view_logs: Optional[bool] = False, save_logs: Optional[bool] = False) -> bool:
        try:
            if token is None:
                token = Utils.get_input("Write The Token » ")
            session = Utils.session_dict(name_session,token,user_agent,time_out,display_welcome,view_logs,save_logs)
            Utils.save_dict(session, f"{name_session}.faru")
            return True
        except:
            return False
    
    @staticmethod
    def open_session(name_session: str, token: Optional[str] = None, user_agent: Optional[str] = None, time_out: Optional[float] = None, display_welcome: bool = False, view_logs: Optional[bool] = False, save_logs: Optional[bool] = False) -> dict:
        """بازکردن سشن

        Args:
            name_session (str): اسم سشن
            token (Optional[str], optional): توکن. Defaults to None.
            user_agent (Optional[str], optional): اطلاعات مرورگر درخواست کننده. Defaults to None.
            time_out (Optional[int], optional): زمان خروج. Defaults to 30.
            display_welcome (bool, optional): خوش آمد گویی. Defaults to False.
            view_logs (Optional[bool], optional): نمایش لاگ ها. Defaults to False.
            save_logs (Optional[bool], optional): ذخیره لاگ ها. Defaults to False.

        Raises:
            CreateSessionError: خطا برای ساخت سشن

        Returns:
            dict: اطلاعات سشن
        """
        path_session = name_session + ".faru"
        if os.path.isfile(path_session):
            with open(path_session, "r", encoding="utf-8") as file:
                encrypted_string = file.read().strip()
            try:
                decrypted = Encryption.de(encrypted_string)
                session_data = json.loads(decrypted)
            except:
                cprint("Error for getting last data session !", Colors.RED)
                creating = False
                while not creating:
                    creating = Utils.create_session(name_session,token,user_agent,time_out,display_welcome,view_logs,save_logs)
                    if not creating:
                        raise CreateSessionError("Can Not Create The Session !")
                session_data = Utils.open_session(name_session,token,user_agent,time_out,display_welcome,view_logs,save_logs)
            current_params = {
                'token': token,
                'user_agent': user_agent,
                'time_out': time_out,
                'display_welcome': display_welcome,
                'setting_logs': {
                    'view': view_logs,
                    'save': save_logs
                }
            }
            for param_name, param_value in current_params.items():
                if param_name in session_data:
                    if param_value is not None and session_data[param_name] != param_value:
                        session_data[param_name] = param_value
        else:
            creating = False
            while not creating:
                creating = Utils.create_session(name_session,token,user_agent,time_out,display_welcome,view_logs,save_logs)
                if not creating:
                    raise CreateSessionError("Can Not Create The Session !")
            session_data = Utils.open_session(name_session,token,user_agent,time_out,display_welcome,view_logs,save_logs)
        return session_data
    

    # Mata Data

    @staticmethod
    def data_format(
        data: dict,
        inline_keypad: Optional[list] = None,
        keypad: Optional[list] = None,
        resize_keyboard: Optional[bool] = True,
        on_time_keyboard: Optional[bool] = False,
        metadata: Optional[list] = None,
        meta_data: Optional[list] = None
    ) -> dict:
        if inline_keypad:
            data["inline_keypad"] = {"rows": inline_keypad}
        if keypad:
            data["chat_keypad"] = {
                "rows": keypad,
                "resize_keyboard": resize_keyboard,
                "on_time_keyboard": on_time_keyboard,
            }
            data["chat_keypad_type"] = "New"
        if metadata:
            data["metadata"] = {"meta_data_parts": metadata}
            if meta_data:
                for meta in meta_data:
                    data["metadata"]["meta_data_parts"].append(meta)
        elif meta_data:
            data["metadata"] = {"meta_data_parts": meta_data}
        return data
    
    # Other

    @staticmethod
    async def d_file(file: Union[str , Path , bytes], file_name: str, network: "Network") -> Dict[str, Tuple[str, Union[bytes, bytearray], str]]:
        if isinstance(file, (bytes, bytearray)):
            d_file = {"file": (file_name, file, "application/octet-stream")}
        else:
            try:
                async with aiofiles.open(file, "rb") as fi:
                    fil = await fi.read()
                    d_file = {"file": (file_name, fil , "application/octet-stream")}
            except:
                file_ = (await network.request(str(file),type_send="GET")).content
                d_file = {"file": (file_name, file_, "application/octet-stream")}
        return d_file
    
    @staticmethod
    def check_data(data: dict) -> bool:
        if data.get("status", "") == "OK":
            return True
        return False

    @staticmethod
    def prefer_first(value1: Optional[str] = None, value2: Optional[str] = None) -> str:
        return value1 if value1 else str(value2)

    @staticmethod
    def get_chat_id_type(chat_id: str):
        if chat_id.startswith("b"):
            return "User"
        elif chat_id.startswith("g"):
            return "Group"
        elif chat_id.startswith("c"):
            return "Channel"
        else:
            raise ValueError("chat id is not found")
